package Abstract;

import java.lang.String;
import java.util.*;
import java.net.*;

/**
* Connection Address class.
* The Address class abstracts necessary information for connection.
* The core idea of the Address class is :
* <UL>
* <LI> Rather than hard-coding on the program, you can specify static connection
*	 data on the file. When an agent is initialized, it will read the Address
*	 file and construct AddressTable. The AddressTable can update connection data
*	 dynamically if necessary.<BR>
*	 You should specify at least one Address data if your agent is an 
*	 application(not an applet) like :<BR>
*	 MyAgent,java.stanford.edu,1234,ServerThread,MyAgent Server<BR>
*	 If your agent wants to communicate with other agent, you should add a line<BR>
*	 OtherAgent,sort.aol.com,3456,ReceiverThread,(Sort Service Agent)<BR>
*	 As you note, the connection data you specified in Address file is static data
*	 in a sense that you should know the other agent name,host, port. 
* </UL>
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class Address {
	/**
	* Agent ID
	*/
	String _id;
	
	/**
	* Internet address(do not support IP address), such as java.stanford.edu
	*/
	String _host;
	
	/**
	* Port number. For Server, you may not specify
	*/
	int _port;
	
	/**
	* Class name for the specific type. The type may represent the Message-Method the 
	* generated ReceiverThread will use.
	*/
	String _type;
	
	/**
	* Description of this Connection Address. Simply this may be used
	* for reference. Or you may use the description has meaning such as<BR>
	* '(SortService :Type sort :Protocol SortProtocol :SubServiceOf Service)'<BR>
	* which show that the description uses KQML and you can add symantics on 
	* the descrpiption.
	*/
	String _description;
	
	/**
	* Default Constructor
	*/
	public Address()  {
		_id = "";
		_host = "";
		_port = -1;
		_type = "";
		_description = "";
	}
	
	/**
	* Constructor given agent name, host, port , type and description.
	* @param id	Name of the Agent
	* @param host Internet address of the agent(IP address is not supported)
	* @param port Port number
	* @param type Normally message-method
	* @param description String description of the Address. Within the description,
	*				 comma(',') is not allowed.
	* @exception none 
	*/
	public Address(String id,String host,int port,String type,String description)  {
		_id = id;
		_host = host;
		_port = port;
		_type = type;
		_description = description;
	}
	
	/**
	* Constructor by reading one line in the Address file
	* You can pass one line read from the Adress file which has the format<BR>
	* ID,Host,port,type,description<BR>
	* The 'description' field is optional.
	* @param addrLine One line read from the Address file
	* @exception ConnectionException Occurs when the input parameter line is not
	* well formated
	*/
	public Address(String addrLine)   
	throws ConnectionException  {
    StringTokenizer st = new StringTokenizer(addrLine,","); 
   	int nofToken = st.countTokens();

   	// should have at least three tokens - id, host, port, type
   	if((nofToken < 4) || (nofToken > 5))  {
     		throw new ConnectionException("ConnectionTable File Format Error");
   	}
   	else  {
   		try  {
     		_id = st.nextToken();
     		_host = st.nextToken();
     		if(_host.equalsIgnoreCase("null"))  {
     			_host = InetAddress.getLocalHost().getHostName();
     		}	
				_port = (Integer.valueOf(st.nextToken())).intValue();
     		_type = st.nextToken();
     		if(st.hasMoreTokens())  {
     			_description = st.nextToken();
     		}
     		else  {	
     			_description = null;
     		}
     	} catch (Exception e)  {
     		System.out.println(toString());
     	}		
 		}	
	}				
	
	/**
	* Set Agent Name
	* @param id Agent Name
	* @return None
	* @exception None
	*/
	public void setID(String id)  {
		_id = id;
	}
	
	/**
	* Set Agent Internet host name
	* @param host Host name
	* @return None
	* @exception None
	*/
	public void setHost(String host)  {
		_host = host;
	}
	
	/** 
	* Set Agent Connection port
	* @param port Port number
	* @return None
	* @exception None
	*/
	public void setPort(int port) {
		_port = port;
	}

	/**
	* Set type
	* @param type Agent type
	*/
	public void setType(String type)  {
		_type = type;
	}
	
	/**
	* Set description
	* @param desc Description
	*/
	public void setDescription(String desc)  {
		_description = desc;
	}
			
	/**
	* Get Agent Name
	* @param None
	* @return _id
	*	@exception None
	*/
	public String getID()  {
		return _id;
	}
	
		/**
	* Get Agent Host Name
	* @param None
	* @return _host
	*	@exception None
	*/
	public String getHost()  {
		return _host;
	}
	
		/**
	* Get Agent Port
	* @param None
	* @return _port
	*	@exception None
	*/
	public int getPort()  {
		return _port;
	}
	
	/**
	* Get an agent type(class name)
	* @return _type of the Address
	*/
	public String getType()  {
		return _type;
	}
		
	/**
	* Get Agent Description
	* @param None
	* @return _description
	*	@exception None
	*/
	public String getDescription()  {
		return _description;
	}	
	
	/** 
	  * toString override. Comma separated string will be returned
	  * @return "id,host,port,type,description"
 	  */
	public String toString()   {
		String str = _id;
		str = str + "," + _host + "," + _port + "," + _type + ",";
		if(_description != null)  {
			str += _description;
		}
			
		return str;
	}	  
}												