package BaseLayer;

import java.lang.*;
import java.util.*;
import java.net.*;
import java.io.*;

import Abstract.*;

/**
* Implementation of AgentAction. This class use BServThread,
* BRecvThread,BSecurity,BMessageBuffer,BConnectionTable,BAddressTable
* as its first implementation from the abstract classes.
* However this class is still an abstract class for you to fill in
* processMessage(String,Object) and Act(Object) methods.
* You will need to subclass this class to act according to the
* message received. Since JATLite assumes to use MessageBuffer,
* the ReceiverThread will receive message and add it to MessageBuffer.
* AgentAction will check the received message and retrieve it 
* whenever available.<BR>
* @see Abstract.AgentAction
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 
public abstract class BAgentAction extends  AgentAction  {
	  
	/**
	* Default constructor. You should perform member-wise initialization for all data 
	* members - setAddressTable(AddressTable),setConnectionTable(ConnectionTable),
	* setDurationTime(int),setEndWith(char), setMessageQueue(MessageBuffer), 
  * setSecurity(Security).
	*/
	public BAgentAction()  {
		super();
	}
	
	/**
	* Constructor with agent name. You should perform member-wise initialization for all data 
	* members - setAddressTable(AddressTable),setConnectionTable(ConnectionTable),
	* setDurationTime(int),setEndWith(char), setMessageQueue(MessageBuffer), 
  * setSecurity(Security).
	* @param id AgentAction(Thread) name
	*/
	public BAgentAction(String id)  {
		super(id);
	}

	/**
	* Constructor with agent name and thread group. You should perform member-wise initialization for all data 
	* members - setAddressTable(AddressTable),setConnectionTable(ConnectionTable),
	* setDurationTime(int),setEndWith(char), setMessageQueue(MessageBuffer), 
  * setSecurity(Security).
	* @param tg Thread group
	* @param id AgentAction(Thread) name
	*/
	public BAgentAction(ThreadGroup tg,String id)  {
		super(tg,id);
	}

	/**
	* Constructor
	* @param id agent(my) name
	* @param queue MessageBuffer
	* @param addresses AddressTable
	* @param connections ConnectionTable
	* @param security Security to be checked and update connection table 
	*/
  public BAgentAction(String id,
		      MessageBuffer queue,
		      AddressTable addresses,
		      ConnectionTable connections,
		      Security security)  {
    super(id,queue,addresses,connections,security);										 
  }
  
  
  /**
   * Constructor. Thread group should passed if you want to access to the ReceiverThread, will
   * be initiated by this action.
   * @param id agent(my) name
   * @param queue MessageBuffer
   * @param addresses AddressTable
   * @param connections ConnectionTable
   * @param security Security to be checked and update connection table 
   * @param tg Thread group, for the receiver thread, initiated by me.
   */
  public BAgentAction(String id,
		      MessageBuffer queue,
		      AddressTable addresses,
		      ConnectionTable connections,
		      Security security,
		      ThreadGroup tg)  {
    super(id,queue,addresses,connections,security,tg);										 
  }
  
  /**
   * Create BServerThread using name of the agent. Endwith character and duration time
   * will be set to those of the current AgentAction's data values.
   * Override this method if your connection type is different.
   * @param id ServerThread name
   * @param priority Server thread priority
   * @return ServerThread, generated. Null if exception occurs.
   */
  public ServerThread createServerThread(String id,int priority)  {
    ServerThread server = null;
    try  {
      server = new BServerThread(_addresses.getAddress(id),_connections,_queue,_security);
      System.out.println("Server created");		
      server.setPriority(priority);
      server.setEndWith(_endWith);
      server.setDurationTime(_durationTime);
      server.start();
      System.out.println(id + "Server Started");		
    } catch (Exception e)  {
      System.out.println(e.toString());
    }	
    
    return server;
  }	
  
  
  /**
   * Create client receiver thread using name of the agent where you want to initiate
   * Connection.
   * Override this method if your connection type is different.
   * @param serverid agent name. This name is not my name. The name of the other
   * agent you want to connect
   * @param myid My (agent)name
   * @param priority Receiver thread priority
   * @exception ConnectionException if security check fails
   * @return ReceiverThread. null if error occurs.
   */
  public synchronized ReceiverThread createReceiverThread(String serverid,String myid,int priority) 
    throws ConnectionException {
      ReceiverThread client = null;
      // Server address should be known
      Address addr = _addresses.getAddress(serverid);
      if(addr == null)  {
	throw new ConnectionException("Server address should be specified. Can not create receiver thread");
      }	
      
      Address myaddr = _addresses.getAddress(myid);
      if(myaddr == null)  {
	throw new ConnectionException("My address should be specified. Can not create receiver thread");
      }
      
      // remove first if already exist, it should be stop first.	
      ReceiverThread rt = _connections.getConnection(myid);
      if(rt != null)  {
	rt.endConn();
      }
      
      try  {
	
	client = new BRecvThread(addr,priority,myaddr,_connections,_queue,_security,getThreadGroup());
	client.setEndWith(_endWith);
	client.setDurationTime(_durationTime);
	// start to receive message
	client.start();		
      }  catch (Exception e)  {
	throw new ConnectionException("Receiver Thread for " + serverid + " can not be created due to " + e.toString());		
      }
      
      return client;
  }	
  
  /**
   * Create receiver thread with AgentAction's name. When you use the method, the
   * AgentAction's thread name(the same as my name) should be set. If not, the receiver
   * thread name will be set as system selected.
   * @param serverid Agent name to whom I want to connect
   * @param priority Priority of the ReceiverThread
   * @exception ConnectionException if security check fails
   * @return ReceiverThread, generated. Null if exception occurs
   */
  public ReceiverThread createReceiverThread(String serverid,int priority) 
    throws ConnectionException {
      return this.createReceiverThread(serverid,getName(),priority);
  }	
  
  /**
   * Send Message by specifying receiver. It first check whether the receiver connection
   * is living. If dead, it try to initiate to the receiver, by retrieving the receiver
   * address from the AddressTable. 
   * @param receiver Receiver Agent
   * @param msg Message to be sent
   * @exception Abstract.ConnectionException If IO error or fail to reconnect, ConectionException will be thrown.
   */
  public void sendMessage(String receiver,String msg) 
  throws ConnectionException   {
    ReceiverThread rthread = _connections.getConnection(receiver);
    if((rthread != null) && (rthread.isAlive())) {
System.out.println("Sending message");    	
      rthread.writeMsg(msg);
System.out.println("Sending success");    	
      
    }
    else  {
      Address address = _addresses.getAddress(receiver);
      if(address == null)   {
	// you may want to request address to ANS though....
	throw new ConnectionException("Can not find address for " + receiver + " Can not reconnect");
      }
      else  {
	// if port = -1, it is registered only as a client
	if(address.getPort() == -1)   {
	  throw new ConnectionException("Connection " + address.getID() + " does not have ServerSocket and it is off-line. Aborted");
	}
	else {	
	  ReceiverThread rt = 
	    createReceiverThread(receiver,Thread.NORM_PRIORITY);
	    
	  if(rt != null)   {
	    rt.writeMsg(msg);	
	  } else  {
	    throw new ConnectionException("Connection to the receiver " + receiver + " failed. Can not send a message");
	  }			
	}	
      }	
    }	
  }
  
  
  /**
   * Broad cast to all the running connecitons
   * @param msg Message to be sent all the living connections
   */
  public void broadCast(String msg)  
  throws ConnectionException   {
    Enumeration e = _connections.elements();
    while(e.hasMoreElements())  {
      ReceiverThread rthread = (ReceiverThread)e.nextElement();
      rthread.writeMsg(msg);
    }
  }
  
  /**
   * Abstract method to process message given any other objects
   * This method is a data exchange between threads or between AgentAction and GUI.
   * @param command Command to be executed
   * @param obj Reference object
   */
  public abstract void processMessage(String command,Object obj);
  
  /**
   * Actual Action method. Note that the parameter is object, which should
   * be instantiated by casting according to the parameter type
   * @param o Message object,received from the message queue. You should cast the
   * message from the object.
   */
  protected abstract boolean Act(Object o);
  
  /**
   * Override run() method as you need. You can check the MessageBuffer
   * and invoke Act(object) if the MessageBuffer is not empty
   */
  public void run()  {
    while(true)  {
      if(!_queue.isEmpty())  {				
	Act(_queue.getMessage());
      }
      else  {
	try  {
	  sleep(100);
	}  catch (InterruptedException e)  {}	
      }
    }
  }				
  
  /**
   * Clean up all the living receiver thread. The living thread will send disconnect
   * message through the connection.
   */
  public void endAction()  {
    try  {
      // send disconnection message to all the connected agents
      Enumeration e = _connections.elements();
      while(e.hasMoreElements())  {
	ReceiverThread rthread = (ReceiverThread)e.nextElement();
	rthread.endConn();
      }
      
      /////////////////////////////////////////////////////
      // If you want to serialize the remaining messages
      // in the _queue, you may need some work here. 
      /////////////////////////////////////////////////////
      _queue.removeAllMessages();
      
      stop();
      join();
    } catch (Exception e)  {
      System.out.println(e.toString());
    }
  }		
}		
		
		
																	 	
