package KQMLLayer;

import java.lang.*;
import java.net.*;
import java.io.*;
import java.util.*;

import Abstract.*;
import BaseLayer.*;

/**
* Extension of BAgentAction, BaseLayer AgentAction class. This class
* is still an abstract class for you to fill in Act(Object) and
* processMessage(String,Object). The received message is KMQLmessage,
* so your Act(Object) method will look like:<P>
* protected void Act(Object o)  {<BR>
*	<UL>String msg = (String)o;<BR>
*			KQMLmessage kqml = new KQMLmessage(msg)<BR>
*			String perf = kqml.getValue("performative");<BR>
*			if(perf.equals(".......	<BR>
*			....</UL>
* }<BR>
* Beta version support full KQML syntax suggested in 1993.<BR>
* To be safe, use sendKQMLMessage(String) or sendMessage(KQML) rather than 
* sendMessage(String receiver,String msg).
* @see Abstract.AgentAction
* @see BaseLayer.BAgentAction
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public abstract class KQMLAgentAction extends BAgentAction  {

  /**
   * Default constructor. You should perform member-wise initialization for all data 
	* members - setAddressTable(AddressTable),setConnectionTable(ConnectionTable),
	* setDurationTime(int),setEndWith(char), setMessageQueue(MessageBuffer), 
  * setSecurity(Security).
   */	
  public KQMLAgentAction()  {
    super();
  }
  
  /**
   * Constructor with Agent name. You should perform member-wise initialization for all data 
	* members - setAddressTable(AddressTable),setConnectionTable(ConnectionTable),
	* setDurationTime(int),setEndWith(char), setMessageQueue(MessageBuffer), 
  * setSecurity(Security).
   * @param id Agent name
   */
  public KQMLAgentAction(String id)   {
    super(id);
  }
  
  /**
   * Constructor with agent name and thread group. You should perform member-wise initialization for all data 
	* members - setAddressTable(AddressTable),setConnectionTable(ConnectionTable),
	* setDurationTime(int),setEndWith(char), setMessageQueue(MessageBuffer), 
  * setSecurity(Security).
   * @param tg Thread group
   * @param id AgentAction(Thread) name
   */
  public KQMLAgentAction(ThreadGroup tg,String id)  {
    super(tg,id);
  }
  
  /**
   * Constructor
   * @param id Agent name
   * @param queue MessageBuffer
   * @param addresses AddressTable
   * @param connections ConnectionTable
   * @param security Security which will be used both for ServerThread and ReceiverTherad
   */
  public KQMLAgentAction(String id,
			 MessageBuffer queue,
			 AddressTable addresses,
			 ConnectionTable connections,
			 Security security)  {
    super(id,queue,addresses,connections,security);										 
  }
  
  /**
   * Constructor
   * @param id Agent name
   * @param queue MessageBuffer
   * @param addresses AddressTable
   * @param connections ConnectionTable
   * @param security Security which will be used both for ServerThread and ReceiverTherad
   * @param tg Thread group
   */
  public KQMLAgentAction(String id,
			 MessageBuffer queue,
			 AddressTable addresses,
			 ConnectionTable connections,
			 Security security,
			 ThreadGroup tg)  {
    super(id,queue,addresses,connections,security,tg);										 
  }
  
  
  /**
   * Create KQMLServerThread using name of the agent. Endwith character and duration time
   * will be set to those of the current AgentAction's data values.
   * Override this method if your connection type is different.
   * @param id ServerThread name
   * @param priority Server thread priority
   * @return ServerThread, generated. Null if exception occurs.
   */
  public ServerThread createServerThread(String id,int priority)  {
    ServerThread server = null;
    try  {
      server = new KQMLServerThread(_addresses.getAddress(id),_connections,_queue,_security);
      System.out.println("Server created");		
      server.setPriority(priority);
      server.setEndWith(_endWith);
      server.setDurationTime(_durationTime);
      server.start();
      System.out.println(id + "Server Started");		
    } catch (Exception e)  {
      System.out.println(e.toString());
    }	
    
    return server;
  }	
  
  /**
   * Create client receiver thread using name of the agent where you want to initiate
   * Connection.
   * Override this method if your connection type is different.
   * @param serverid agent name. This name is not my name. The name of the other
   * agent you want to connect
   * @param myid My (agent)name
   * @param priority Receiver thread priority
   * @return ReceiverThread. null if error occurs.
   */
  public synchronized ReceiverThread createReceiverThread(String serverid,String myid,int priority) 
    throws ConnectionException {
      
      ReceiverThread client = null;
      // Server address should be known
      Address addr = _addresses.getAddress(serverid);
      if(addr == null)  {
	System.out.println("Server address should be specified");
	throw new ConnectionException("Server address should be specified");
      }	
      
      Address myaddr = _addresses.getAddress(myid);
      if(myaddr == null)  {
	System.out.println("My address should be specified");
	throw new ConnectionException("My address should be specified");
      }
      
      // remove first if already exist, it should be stop first.	
      ReceiverThread rt = _connections.getConnection(myid);
      if(rt != null)  {
	rt.endConn();
      }
      
      try  {
	System.out.println("Before");
	
	client = new KQMLRecvThread(addr,priority,myaddr,_connections,_queue,_security,getThreadGroup());
	System.out.println("after");
	
	client.setEndWith(_endWith);
	client.setDurationTime(_durationTime);
	// start to receive message
	client.start();		
      }  catch (Exception e)  {
	System.out.println("Receiver Thread for " + serverid + " can not be created due to " + e.toString());
	throw new ConnectionException("Connection fails " + e.toString());
      }
      
      return client;
  }	
  
  /**
   * Create receiver thread with AgentAction's name. When you use the method, the
   * AgentAction's thread name(the same as my name) should be set. If not, the receiver
   * thread name will be set as system selected.
   * @param serverid Agent name to whom I want to connect
   * @param priority Priority of the ReceiverThread
   * @return ReceiverThread, generated. Null if exception occurs
   */
  public ReceiverThread createReceiverThread(String serverid,int priority) 
    throws ConnectionException {
      return this.createReceiverThread(serverid,getName(),priority);
  }	
  
  
  /**
   * Send KQML object by converting to String
   * @param kqml KQMLmessage object
   * @exception Abstract.ConnectionException
   */
  public void sendMessage(KQMLmessage kqml) 
  throws ConnectionException   {
    sendMessage(kqml.getValue("receiver"),kqml.getSendString());
  }

  /**
   * Send KQMLmessage. Before it to be sent, parse error will be checked
   * and throw ParseException if not KQML syntax.
   * @param KQMLstring KQML syntax message string
   * @exception KQMLLayer.ParseException, Abstract.ConnectionException
   */
  public void sendKQMLMessag(String KQMLstring)  
    throws ParseException, ConnectionException {
      KQMLmessage kqml = new KQMLmessage(KQMLstring);
      sendMessage(kqml);
  }
} 		
		
	
		
	