package KQMLLayer;

import java.io.*;
import java.util.*;

  /** KQMLStreamTokenizer is a scanner for KQML-messages. it is
      api-compatible with JDK's StreamTokenizer but rewritten to
      support multi-line-strings as specified in the KQML-syntax.
  */
/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <mailto:raquet@cdr.stanford.edu>Andreas Raquet</A>
*  @version Java(tm) 1.1
*/

public class KQMLStreamTokenizer  {
  protected Reader reader;
  protected int syntax[];
  protected boolean pushed;
  public int nval,ttype, lastChar;
  public String sval;
  public static int TT_WORD=1;
  public static int TT_NUMBER=2;
  public static int TT_EOF=3;
  public static int TT_EOL=4;
  /** The constructor for KQMLStreamTokenizer requires a java.io.Reader
      to supply the incoming character-stream.
      @param r incoming character-stream
  */
  public KQMLStreamTokenizer(Reader r) {
    reader=r;
    resetSyntax();
    wordChars('\u0021','\u00FF');
    whitespaceChars('\u0000','\u0020');
    ordinaryChar('(');
    ordinaryChar(')');
    ordinaryChar('\'');
    ordinaryChar(',');
    quoteChar('"');
    fixedQuoteChar('#');
    lastChar=-1;
  }
  /** Clear the internal syntax-table
   */
  protected void resetSyntax() {
    syntax=new int[256];
  } 
  /** Specifies that all characters between low and high are considered
      word-characters. The specified interval of characters are added to the
      set of word-characters, old word-characters remain in place.
      @param low lower bound of the character-interval
      @param high upper bound of the character-interval
  */
  public void wordChars(char low, char high) {
    for (int i=low; i<=high; i++) syntax[i]=1;}
  /** Specifies that all characters between low and high are considered
      whitespace-characters. The specified interval of characters are added to the
      set of whitespace-characters, old whitespace-characters remain in place.
      Whitespace delimits words, but is not returned as tokens.
      @param low lower bound of the character-interval
      @param high upper bound of the character-interval
  */
  public void whitespaceChars(char low, char high) {
    for (int i=low; i<=high; i++) syntax[i]=0;}
  /** Specifies that the given character to be an ordinary character. 
      Ordinary characters delimit words and are returned as separate tokens
      @param c new ordinary character
  */
  public void ordinaryChar(char c) { 
    syntax[(int)c]=2; }
  /** Specifies that the given character to be a quote character.
      Quote characters mark the beginning and end of a string.
      Inside a string quote-characters must be escaped by a \
      @param c new quote character
  */
  public void quoteChar(char c) { 
    syntax[(int)c]=3; }
  /** Specifies that the given character to be a fixed-quote character.
      Fixed-Quote characters mark the beginning of a fixed string.
      After the fixed-quote-character, a number must state the fixed
      length of the string. String-delimiters are ignored in fixed strings.
      @param c new fixed-quote character
  */
  public void fixedQuoteChar(char c) { 
    syntax[(int)c]=4; }
  /** pushes back the last token into the incoming-stream. The next read-
      activity will return the same token again.
  */
  public void pushBack() { pushed=true; }
  /** Reads the next token from the incoming stream and returns the
      type-id of that token. Usually getNextToken() is used to directly return
      the token instead of only the type. nextToken is a more efficient way
      to skip tokens if one is not interested in the token.
  */
  public int nextToken() throws ParseException,IOException {
    if (pushed) {
      pushed=false;
      return 0;}
    int c;
    if (lastChar>=0)
      c =lastChar;
    else c=reader.read();
    if (c==-1) {
      ttype=TT_EOF;
      return 0;}
    ttype = syntax[c];
    while (ttype==0) {
	c=reader.read();
	if (c==-1) ttype=TT_EOF;
	else ttype=syntax[c];}
    switch (ttype) {
    case 2:
      ttype = c;
      sval=null;
      lastChar=reader.read();
      break;
    case 1:
      StringBuffer w = new StringBuffer(64);
      while (ttype==1) {
	w.append((char)c);
	c=reader.read();
	if (c==-1) ttype=TT_EOF;
	else ttype=syntax[c];}
      sval=w.toString();
      ttype=TT_WORD;
      lastChar=c;
      break;
    case 3:
      StringBuffer s = new StringBuffer(64);
      try {
	c=reader.read(); // get first char after "
	if (c==-1) ttype=TT_EOF;
	else ttype=syntax[c];
	while (ttype!=3) {
	  if (c=='\\') c=reader.read();
	  s.append((char)c);
	  c=reader.read();
	  if (c==-1) ttype=TT_EOF;
	  else ttype=syntax[c];}
	sval = s.toString();
	ttype='"';
	lastChar=reader.read();}
      catch (ArrayIndexOutOfBoundsException e) {
	throw new ParseException("String: missing closing quote"); }
      break;
    case 4:
      StringBuffer fs = new StringBuffer(64);
      try {
	c=reader.read();
	if (c==-1) ttype=TT_EOF;
	else ttype=syntax[c];
	while (ttype==1) {
	  fs.append((char)c);
	  c=reader.read();
	  if (c==-1) ttype=TT_EOF;
	  else ttype=syntax[c];}
	lastChar=c;
	int len = Integer.valueOf(fs.toString()).intValue();;
	char[] in = new char[len+1];
	int res=reader.read(in,0,len);
	if (res!=len) throw new ParseException("FixedString: unexpected EOF"); 
	sval = new String(in);
	ttype='#';
	lastChar=reader.read();
      }
      catch (NumberFormatException e) {
	throw new ParseException("Expected: FIXED_LEN_STRING Found: NO_LENGTH_GIVEN");}
      catch (ArrayIndexOutOfBoundsException e) {
	throw new ParseException("FixedString: unexpected EOF"); }
      break;
    }
    return 0;
  }
  /** A convenience-method that reads the next token, compares it with the
      given string using .equalsIgnoreCase() and throws a ParseException if they
      do not match.
      @param expected expected token
  */
  public void consumeIgnoreCase(String expected) throws IOException, ParseException {
    String found=getNextToken();
    if (!expected.equalsIgnoreCase(found)) 
      throw new ParseException("Expected: "+expected+" Found: "+found);
  }
  /** A convenience-method that reads the next token, compares it with the
      given string using .equals() and throws a ParseException if they
      do not match.
      @param expected expected token
  */
  public void consume(String expected) throws IOException, ParseException  {
    String found=getNextToken();
    if (!expected.equals(found)) 
      throw new ParseException("Expected: "+expected+" Found: "+found);
  }
  /** A convenience-method that reads the next token, compares it with the
      given string using .equals() and throws a ParseException if they
      do not match or the token is not a single character.
      @param expected expected token
  */
  public void consume(char expected) throws IOException, ParseException  {
    nextToken();
    if (ttype==TT_WORD) {
      if (sval.length()==1 && sval.charAt(0)==expected) return;
    }
    else if ((char)ttype==expected) return;
    throw new ParseException("Expected: "+String.valueOf(expected)+" Found: "+String.valueOf(ttype));
  }
  /** Returns the next token from the input-stream without removing it (aka
      "lookahead"). Thus
      multiple calls to viewToken() will always return the same token.
  */
  public String viewToken() throws IOException, ParseException {
    String ret = getNextToken();
    pushBack();
    return ret;}
  /** A convenience-method that reads the next token and compares it with the
      given character. It returns true if they match,
      false if they don't or if the next token is not a single
      character.
      @param expected token to compare to the input
  */  
  public boolean testToken(char expected) throws IOException, ParseException{
    nextToken();
    pushBack();
    if (ttype==TT_WORD)
      return (String.valueOf(expected).equals(sval));
    else
      return (ttype==expected);}
  /** A convenience-method that reads the next token and compares it with the
      given string using .equals(). It returns true if they match,
      false otherise.
      @param expected token to compare to the input
  */  
  public boolean testToken(String expected) throws IOException, ParseException{
    nextToken();
    pushBack();
    return (ttype==TT_WORD) && (sval==expected);}
  /** returns the next token from the input-stream. The token is always
      returned as a string.
  */
  public String getNextToken() throws IOException, ParseException  {
    nextToken();
    if (ttype==TT_WORD) return sval;
    else if ((char)ttype=='"') return sval;
    else if ((char)ttype=='#') return sval;
    else return String.valueOf((char)ttype); }
}