package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;

import java.net.*;
import java.io.*;

/**
 * Socket connector between an applet and server(SMTP,POP3,FTP,HTTP,etc.). Since an applet can
 * not create socket other than it spawned from, router will take a role as proxy to 
 * connect the applet and servers. Router will create connector which connects the applet and
 * servers. Connector will have two sockets, one for a server and one for an applet. Data input
 * from any of two socket will be redirected to other side. Initially, a server address(host,port)
 * will be passed to the constructor. After Connector connect to the server, it will create a
 * server socket for the applet. The server socket port number will be passed to the applet and the applet will
 * connect to the server socket. The socket port number will be passed as high level KQML message
 * with performative 'port-info'.
 */
 
/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 
 
public class ConnectorThread extends Thread  {
	
	/*
	 * Receiver thread for a client applet
	 */
	protected BRecvThread _client;
	
	/**
	 * Message buffer for a client applet
	 */
	protected BMessageBuffer _clientQueue;
	
	/**
	 * Receiver thread for an internet server such as SMTP, FTP
	 */
	protected BRecvThread _server;
	
	/**
	  * Message buffer for a server thread
	  */
	protected BMessageBuffer _serverQueue;
	
	/**
	  * Server socket, created for an applet connection for data transfer
	  */
	protected ServerSocket _clientServerSocket;
	
	/**
	  * Maximum time set in minutes
	  */
	protected int _maxWaitTime;
	
	/**
	  * Message endWith character. Default is '\n'.  You need to use setEndWith(char) method 
	  * if messages are not ended with '\n'.
	  */
	protected char _endWith;
	
	/**
	  * Timer to check time out. If time out, the thread will stop this thread and timer itself
	  * will stop. If data transfer is successful, timer will be stop by this thread.
	  */
	protected ConnectorTimerThread _timer;
	
	/**
	  * Default constructor
	  */
	public ConnectorThread()  {
		super();
	}
	
	
	/**
	  * Constructor with internet service server address, priority and maximum idle time.
	  * Router will try to connect to given service address.
	  * @param serverAddress Internet service server address.
	  * @param priority Thread priority
	  * @param connectionWaitTime Maximum idle time.
	  * @exception Abstract.ConnectionException If server connection is failed, or 
	  * fails to create ServerSocket for an applet.
	  */	
	public ConnectorThread(Address serverAddress,
												 int priority,
												 int connectionWaitTime) throws ConnectionException  {
		try  {
			Socket csocket = new Socket(serverAddress.getHost(),serverAddress.getPort());
			init(csocket,priority,connectionWaitTime);
		}	catch (Exception e)  {
			throw new ConnectionException(e.toString());
		}	
	}	
	
	/**
	  * Constructor with internet service server address, priority and maximum idle time.
	  * Router will try to connect to given service address.
	  * @param toSocket Internet service server socket.
	  * @param priority Thread priority
	  * @param connectionWaitTime Maximum idle time.
	  * @exception Abstract.ConnectionException If server connection is failed, or 
	  * fails to create ServerSocket for an applet.
	  */	
	public ConnectorThread(Socket toSocket,
												 int priority,
												 int connectionWaitTime)  throws ConnectionException {
		try  {												 
			init(toSocket,priority,connectionWaitTime);
		} catch (Exception e)  {
			throw new ConnectionException(e.toString());
		}		
	}	
	
	/**
	  * Initialize data members, Server thread will be created and ServerSocket for an applet 
	  * will be created. Client thread for an applet will not be initialized here. Initialization
	  * of client thread will be performed at run time, after accept a connection from the applet.
	  * @param socket Socket for internet service server
	  * @param priority Priority for this thread
	  * @param connectionTime Max. idle time 
	  * @exception Abstract.ConnectionException 
	  */
	protected void init(Socket socket,int priority,int connectionTime) 
		throws ConnectionException {
		try  {
			_serverQueue = new BMessageBuffer();
			_server = new BRecvThread(socket,priority,null,null,_serverQueue,null);
			_client = null;
			_clientQueue = null;
			setEndWith('\n');  // normally '\n'
			_maxWaitTime = connectionTime;		
			_clientServerSocket = new ServerSocket(0);	 
			setPriority(priority);
		} catch (Exception e)  {
			throw new ConnectionException(e.toString());
		}		
	}
		
	/**
	  * Accessor to see whether client applet is connected or not
	  * @return true if connected
	  */		
	public boolean isConnected()   {
		if(_client != null)  {
			return true;
		}
		return false;
	}		
			
	/**
	  * Set endWith character if different from '\n'
	  * @param endWith End with character
	  */	
	public void setEndWith(char endwith)  {
		if(_client != null)  {
			_client.setEndWith(endwith);
		}
		
		if(_server != null)  {
			_server.setEndWith(endwith);
		}
		_endWith = endwith;		
	}
			
	/**
	  * Accessor to get machine generated client ServerSocket number. 
	  * @return ServerSocket number for the applet to connect
	  */
	public int getClientServerSocketNumber()   {
		return _clientServerSocket.getLocalPort();
	}
		
		
	/**
	* Default behavior is to read from any of message queues - server side and client side. 
	* If there's a message, send it to
	* counter part(e.g. a message from server side will send to applet , vice versa.)
	* Default behavior may not be appropriate for FTP, HTTP, POP3, etc. For each protocol,
	* Connector will need to be extended. Timer will be set for waiting connection from an
	* applet. If time limit is over, thread will close all the connection and stop.
	*/		
	public void run()   {
		// timer set for unstable applet condition
		resetTimer();
					
    try {
			Socket client_socket = _clientServerSocket.accept();
			_clientQueue = new BMessageBuffer();
			_client = new BRecvThread(client_socket,getPriority(),null,null,_clientQueue,null);
			_client.setEndWith(_endWith);  // normally '\n'
			_clientServerSocket.close();
			_server.start();
			_client.start();

			while(true)  {
				if(!_clientQueue.isEmpty())  {	
				
					String msg = (String)_clientQueue.getMessage();
					_server.writeMsg(msg);
					
					msg = msg.toLowerCase();
					// normally, quit is end condition
					if(msg.startsWith("quit"))  {
						endConn();
					}
					else  {				
						resetTimer();
					}	
				}
				else
				if(!_serverQueue.isEmpty())  {
					_client.writeMsg((String)_serverQueue.getMessage());
					resetTimer();
				}	
				else  {
					try  {
						sleep(100);
					}  catch (InterruptedException e)  {}	
				}
			}
		}  catch (Exception e)  {
			System.out.println(e.toString());
			endConn();
		}	
	}	
	
	/**
	  * Disconnect connection but does not kill timer. This method should be invoked only from
	  * ConnectorTimer.
	  */
	public void disconnectCon()  {
		if(_client != null)  {
			_client.endConn();
			_client = null;
		}
		if(_server != null)  {
			_server.endConn();
			_server = null;
		}
		
		
		if(_clientServerSocket != null)  {
			try  {
				_clientServerSocket.close();
				_clientServerSocket = null;
			} catch (IOException e)  {
				System.out.println(e.toString());
			}		
		}	
	
		if(isAlive())  {
System.out.println(getName() + " stop");		
    	stop();
    	try  {
      	join();
    	} catch (InterruptedException e3)  {}	
    }	
		
	}	
	
	/**
	  * Kill timer and disconnect.
	  */
	protected void endConn()  {
//System.out.println("Connector stop");	
	
		stopTimer();
		disconnectCon();
	}
		
	/**
	  * Stop timer thread
	  */	
	protected void stopTimer()  {
		if(_timer != null)  {
			if(_timer.isAlive())  {
System.out.println("Timer stop" + _timer.getName());				
				_timer.stop();
				_timer = null;
			}
		}
	}	
	
	/**
	  * Reset timer. If timer is not created yet, created one.
	  */
	protected void resetTimer()  {
		if(_maxWaitTime == -1)  {
			return;
		}
			
		if(_timer != null)   {
			_timer.resetTime();
		}
		else  {		
			_timer = new ConnectorTimerThread(this,_maxWaitTime);
			_timer.start();
		}
	}			
}