package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;
import RouterLayer.util.*;
import RouterLayer.Router.*;

import java.net.*;
import java.io.*;
import java.util.*;

/**
* IPRouter server action class. Add on Router server action, diverse internet protocols - 
* SMTP and FTP are supported.
* @see RouterLayer.Router.RouterAction
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class IPRouterAction extends RouterAction  {

	/**
	* Default constructor. You should use setAddressFilePath, setMessageBoxPath, setPasswordFilePath
	* methods afterwards.
	*/
	public IPRouterAction()  {
	}

	/**
	* Constructor using 
	* @param id Router name
	* @param addressFilePath Address file path. Address file contains the addresses of Router and 
	*  RouterRegistrar at the minimum. Add on those, you can specify the addresses of other agents
	*  where the Router want to initiate the conneciton
	* @param messageBoxPath Message incoming file directory path
	* @param passwordFilePath Password file path
	* @param registryFilePath Registry file path
	* @param duration Maximum idle time
	* @param offLineTrialTimes Maximum trial times to voluntarily reconnect to stand alone agent
	* @param offLineCheckPeriod Off line agent time check period in minutes
	* @param reserveTrialTimes Maximum trial times to send reserved message 
	* @param reserveCheckPeriod Reseved message sending time check period in minutes
	* @param sleepPeriod Sleep period in minutes
	*/
	public IPRouterAction(String id,
						String addressFilePath,
						String messageBoxPath,
						String passwordFilePath,
						String registryFilePath,
						int durationTime,
						int offLineTrialTimes,
						int offLineCheckPeriod,
						int reserveTrialTimes,
						int reserveCheckPeriod,
						int sleepPeriod)  {
		super(id,
					addressFilePath,
					messageBoxPath,
					passwordFilePath,
					registryFilePath,
					durationTime,
					offLineTrialTimes,
					offLineCheckPeriod,
					reserveTrialTimes,
					reserveCheckPeriod,
					sleepPeriod);
	}

	/**
	* Override AgentAction createServerThread method
	* Create server thread using name of the agent
	* Override this method if your connection type is different.
	* @param id ServerThread name
	* @param priority Server thread priority
	* @return ServerThread, generated. Null if fails
	*/
	public ServerThread createServerThread(String id,int priority)  {
		ServerThread server = null;
		try  {
			server =
				new IPRouterServerThread(_addresses.getAddress(id),_connections,_security,_messageBoxPath,_reservedBook,_offLineList,_disconnectTable);
			server.setPriority(priority);
			server.setDurationTime(_durationTime);
			server.setEndWith(_endWith);
			server.start();
		} catch (Exception e)  {
			System.out.println(e.toString());
		}

		return server;
	}


	/**
	* Override AgentAction createRecvThread method
	* Create client receiver thread using name of the agent
	* Override this method if your connection type is different.
	* Before you invoke this method, the Router name(the same as RouterAction thread name)
	* should be specified.
	* @param serverid Server agent name. This name is not my name. The name of the other
	* agent you want to connect
	* @param priority Receiver thread priority
	* @return ReceiverThread, generated
	*/
	public ReceiverThread createReceiverThread(String serverid,int priority)  {
		ReceiverThread client = null;
		// Server address should be known
		Address addr = _addresses.getAddress(serverid);
		if(addr == null)  {
			System.out.println("Client Address for " + serverid + " is not specified");
			return null;
		}
		try  {
			client = new IPRouterRecvThread(addr,priority,_addresses.getAddress(getName()),_connections,new BMessageBuffer(),_security,_messageBoxPath,_reservedBook,_offLineList,_disconnectTable,getThreadGroup());

			client.setEndWith(_endWith);
			client.setDurationTime(_durationTime);
			client.setName(addr.getID());
			client.start();
			System.out.println("Client to " + serverid + " started");
		} catch (Exception e)  {
			System.out.println(e.toString());
		}

		return client;
	}


	public static void main(String args[])  {
		
		if(args.length != 1)  {
			System.out.println("Wrong arguments. Script file should be specified");
			System.exit(0);
		}
		
		try  {
			BufferedReader in = new BufferedReader(new FileReader(new File(args[0])));
			ReadOneLine input = new ReadOneLine(in);

			String name = input.getOneLine();
			String addrpath = input.getOneLine();
			String incomingpath = input.getOneLine();
			String passwordpath = null;
      String registryFilePath = null;
			String password = input.getOneLine();
      String registry = input.getOneLine();
			
			if(!password.equals("null"))  {
				passwordpath = password;
			}
      if(!registry.equals("null"))  {
      	registryFilePath = registry;
      }
			
			int durationtime = Integer.valueOf(input.getOneLine()).intValue();
			int offlinemaxtimes = Integer.valueOf(input.getOneLine()).intValue();
			int offlinechecktime = Integer.valueOf(input.getOneLine()).intValue();
			int reservemaxtimes = Integer.valueOf(input.getOneLine()).intValue();
			int reservechecktime = Integer.valueOf(input.getOneLine()).intValue();
			int sleeptime = Integer.valueOf(input.getOneLine()).intValue();
			
			in.close();
			
			IPRouterAction action = new IPRouterAction(name,
												 										 addrpath,
												 										 incomingpath,
												 										 passwordpath,
												 										 registryFilePath,
												 										 durationtime,
												 										 offlinemaxtimes,
												 										 offlinechecktime,
												 										 reservemaxtimes,
												 										 reservechecktime,
												 										 sleeptime);
			action.start();
		}  catch (Exception e)  {
				System.out.println(e.toString());
		}
	}
}