package ProtocolLayer;

import Abstract.*;
import BaseLayer.*;
import KQMLLayer.*;

import java.util.*;
import java.text.*;
import java.net.*;

/**
* Example class of extending IPThread(Internet protocol thread).
*/
/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/
public class SMTPThread extends IPThread   {
	
	/**
	  * SMTPCon data 
	  */
	SMTPCon _smtpCon = null;


	/**
	* Default Constructor.
	*/
	public SMTPThread()  {
		super();
		_endWith = '\n';
		_cmdList = null;
	}
		
	/**
	* Create ReceiverThread with specific thread name
	* @param name Agent name
	*/	
	public SMTPThread(String name)  {
		super(name);
		_endWith = '\n';
		_cmdList = null;
		
	}
		
	/**
	* Create ReceiverThread with specific thread name and Thread group.
	* Passing thread group is essential when you will create the receiver thread for an applet
	* because of security issue.
	* @param tg Thread group to which this thread will be owned to
	* @param name Agent name
	*/	
	public SMTPThread(ThreadGroup tg,String name)  {
		super(tg,name);
		_endWith = '\n';
	}
		
	/**
	* Can be used both for applets and stand-alone agents, but for efficiency, normally
	* used for applets. Router will create a connector which connects an applet and SMTP
	* server. Through the connector, email will be sent as if the applet is a stand-aloone
	* agent.
	* @param con SMTPCon object which contains proxy server(router) internet address and
	*  connector port, my email address, receiver email address, email data and command list
	* @param priority Thread priority. Default is NORM_PRIORITY
	* @param tg Thread group, the thread will be grouped.
	* @exception Throws ConnectionException if connection error occurs.
	*/	
	public SMTPThread(SMTPCon con,int priority,ThreadGroup tg) throws ConnectionException {
		this(con.getServerAddress(),
				 priority,
				 con.getMyAddress(),
				 con.getReceiverAddresses(),
				 con.getData(),
				 tg);
	}			 	

  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using Address. 'security' object will be used to 
   *  satisfy connection protocol to server. This constructor will be called when
   *  an agent use JATLite router as a proxy. Do not use this constructor when the agent
   *  is a stand-alone agent which has capability to connect email server directly.
   *  For serveraddress parameter, host should be the router host name and port should
   *  be a port, specified by the router. If you need to more advanced command list
   *  other than provided by default, you can pass command list as a parameter, If it is null,
   *  default simple command list will be assumed.
   *  @param serveraddress The address of the remote host(Proxy address, not an actual email server).
   *  @param priority The priority of the thread.
   *  @param myEmail My email address. 
   *  @param receiverEmail Receiver email address.
   *  @param data Email content.
   *  @param cmdlist Command list. If null, default simple command list will be applied
   *  @param tg Thread group
   * @exception ConnectionException if error occurs
   */
  public SMTPThread(Address serveraddress,
		     int priority, 
		     String myEmail,
		     String receiverEmail,
		     String data,
		     ThreadGroup tg)
    throws ConnectionException   {
      this(myEmail,receiverEmail,data,tg);  
      setPriority(priority);  
			
			_proxyAddress = serveraddress;	
			_useProxy = true;
  }
	
  /**
   *  This constructor is called when an agent want to initialize a connection
   *  to the other agents, using Address. 'security' object will be used to 
   *  satisfy connection protocol to server. This constructor will be called when
   *  an agent use JATLite router as a proxy. Do not use this constructor when the agent
   *  is a stand-alone agent which has capability to connect email server directly.
   *  For serveraddress parameter, host should be the router host name and port should
   *  be a port, specified by the router. If you need to more advanced command list
   *  other than provided by default, you can pass command list as a parameter, If it is null,
   *  @param serveraddress The address of the remote host. Host will be the same as router,
   *  and port will be given from the router. 
   *  @param priority The priority of the thread.
   *  @param myaddress My address. 
   *  @param connections the connection table is passed as an argument.
   *  @param queue the Message Queue.
   *  @param security The security object 
   * @exception ConnectionException if error occurs
   */
  public SMTPThread(Address serveraddress,
		     int priority, 
		     String myEmail,
		     String[] receiverEmails,
		     String data,
		     ThreadGroup tg)
    throws ConnectionException   {
      this(myEmail,receiverEmails,data,tg);  
      setPriority(priority);  
			
			_smtpCon = new SMTPCon(serveraddress,myEmail,receiverEmails,data);
			
			_proxyAddress = serveraddress;	
			_useProxy = true;
  }
	

	/**
	*  For simplicity. Default SMTP server port is 25. Command list should be set afterwards.
	* @param myEmailAddress Sender email address - username@host
	* @param tg Thread group
	*/
	public SMTPThread(String myEmailAddress,ThreadGroup tg) throws ConnectionException {
		super(tg,"SMTP" + myEmailAddress);
		try  {
			_myaddress = new Address("SMTPClient",null,-1,"SMTP","(agent-info :email " + myEmailAddress + ")");
			int atpos = myEmailAddress.lastIndexOf("@");
			String host = myEmailAddress.substring(atpos + 1);
			_serverAddress = new Address("SMTPServer",host,25,"SMTPServer","");
			_security = new SMTPSecurity();
			_cmdList = null;
			_smtpCon = new SMTPCon(_serverAddress,myEmailAddress,null,null);
			setEndWith('\n');
		} catch (Exception e)  {
			throw new ConnectionException(e.toString());
		}		
	}

	/**
	*  For simplicity. Default SMTP server port is 25.
	* @param myEmailAddress Sender email address - username@host
	* @param cmdlist Command list	
	* @param tg Thread group
	*/
	public SMTPThread(String myEmailAddress,Vector cmdlist,ThreadGroup tg) 
		throws ConnectionException {
		this(myEmailAddress,tg);
		_cmdList = cmdlist;
	}

	/**
	* Constructor, performing the simplest mail transfer. Mail content will be passed as 'mail' 
	* parameter with receiver email address(receiverMailAddress), my user name(username) and 
	* my mail host(mailhost).
	* @param myEmailAddress Sender email address - username@host
	* @param receiverMailAddress Full receiver email address, e.g. jhc@cdr.stanford.edu
	* @param mail Mail content
	* @param tg Thread group
	*/
	public SMTPThread(String myEmailAddress,
										String receiverMailAddress,
										String mail,
										ThreadGroup tg) throws ConnectionException {
										
		this(myEmailAddress,tg);
		String[] tmp = new String[1];
		tmp[0] = receiverMailAddress;
		_smtpCon.setReceiverAddresses(tmp);
		_smtpCon.setData(mail);
		makeSimpleCommandList();		
	}

	/**
	* Constructor, performing the simplest mail transfer. Parameters are address object based.
	* @param myAddress My address, which should contains my email address in description field
	* @param receiverAddress Receiver address, which should contains receiver email address
	* @param mail Mail content
	* @param tg Thread group
	*/
	public SMTPThread(Address myAddress,
										Address receiverAddress,
										String mail,
										ThreadGroup tg) throws ConnectionException {
		
		if((myAddress == null) || (receiverAddress == null))   {
			throw new ConnectionException("Email address is missing");
		}
		
		String myEmail = getEmailAddress(myAddress);
		if(myEmail.equals(""))  {
			throw new ConnectionException("My email address is not specified in my address");
		}
	
		String rEmail = getEmailAddress(receiverAddress);
		if(rEmail.equals(""))  {
			throw new ConnectionException("Receiver email address is not specified in receiver address");
		}

		try  {
			_myaddress = myAddress;
			int atpos = myEmail.lastIndexOf("@");
			String host = myEmail.substring(atpos + 1);
			_serverAddress = new Address("SMTPServer",host,25,"SMTPServer","");
			_security = new SMTPSecurity();
			setEndWith('\n');
	
			String[] tmp = new String[1];
			tmp[0] = rEmail;
			_smtpCon = new SMTPCon(_serverAddress,myEmail,tmp,mail);
			makeSimpleCommandList();		
		} catch (Exception e)  {
			throw new ConnectionException(e.toString());
		}		
		
	}

	/**
	* Constructor, performing the simplest mail transfer. Mail content will be passed as 'mail' 
	* parameter with receiver email address(receiverMailAddress), my user name(username) and 
	* my mail host(mailhost).
	* @param myEmailAddress Sender email address - username@host
	* @param receiverMailAddressList Full receiver email addresses - multiple receivers, 
	* @param mail Mail content
	* @param tg Thread group
	*/
	public SMTPThread(String myEmailAddress,
										String[] receiverMailAddressList,
										String mail,
										ThreadGroup tg) throws ConnectionException {
										
		this(myEmailAddress,tg);
		_smtpCon.setReceiverAddresses(receiverMailAddressList);
		_smtpCon.setData(mail);
		makeSimpleCommandList();
	}

	/**
	* Make command list from given sender, receiver and mail.
	* Mail header will include extended slot 'X-MessageType' as KQML.
	*
	* Header will be:
	* Date: date
	* Sender: sender
	* Receiver: ReceiverA ReceiverB ReceiverC
	* Subject: KQMLMessage
	* X-MessageType: KQML
	*
	* @param senderAddr Sender address
	* @param receiverAddrs Receiver addresses
	* @param data Data to be sent, normally a KQML message
	*/
	public void makeSimpleCommandList() throws ConnectionException  {
		if(_smtpCon == null)  {
			throw new ConnectionException("SMTP Con not specified. Can not make command list");
		}
			
		String senderAddr = _smtpCon.getMyAddress();
		String[] receivers = _smtpCon.getReceiverAddresses();
		String data = _smtpCon.getData();
		
		if(senderAddr == null)   {
 			throw new ConnectionException("SMTP Sender not specified");
 		}
 			
		if(receivers == null)   {
 			throw new ConnectionException("SMTP Receiver not specified");
		}
 			
		if(data == null)   {
 			throw new ConnectionException("SMTP Data not specified");
 		}
 			
		_cmdList = new Vector();
		
 		SimpleDateFormat df = new SimpleDateFormat("MMM.dd.yyyyy hh:mm aaa");
 		String newdata = "Date: " + df.format(new Date()) + "\n";
 		newdata = newdata + "Sender: " + senderAddr + "\n";
		
		String mailfrom = "MAIL FROM: <" + senderAddr + ">";
		_cmdList.addElement(mailfrom);

		newdata = newdata + "Receiver: ";
		
		for(int i=0; i < receivers.length; i++)  {
			String mailto = "RCPT TO: <" + receivers[i] + ">";
			_cmdList.addElement(mailto);
			newdata = newdata + receivers[i] + ",";
		}
		newdata = newdata + "\n";
		newdata = newdata + "Subject: KQMLMessage\n";
		newdata = newdata + "X-MessageType: KQML\n";
		
		String datacmd = "DATA";
		_cmdList.addElement(datacmd);
		_cmdList.addElement(newdata + data + "\n.");

	}
	
	/**
	  * Get email address from Address object
	  * @param addr Address
	  * @return null if no email is specified. Email address(e.g. jhc@cdr.stanford.edu)
	  * if addr contains email address information.
	  */		
	public String getEmailAddress(Address addr)   {
		if(addr == null)  {
			return "";
		}
		
		String descript = addr.getDescription();
		if(descript == null)  {
			return "";
		}
		
		try  {
			KQMLmessage kqml = new KQMLmessage(descript);
			return kqml.getValue("email");
		}  catch (Exception e)  {
			notifyError("Address description field syngax error");
		}
		
		return "";
	}	
						
	/**
	* processCommands implementation. Send predefined command to mail server and manage received
	* response.
	*/
	public void processCommands()   {
		boolean isError = false;
		
		try  {
			if(_cmdList != null)   {
				Enumeration e = _cmdList.elements();
				while(e.hasMoreElements())  {
					String cmd = (String)e.nextElement();
					writeMsg(cmd);
					String returnmsg = getOneMessage();
					
					// if AgentAction want to get control, give it a control
					if(_ipAgentAction != null)  {
						_ipAgentAction.processMessage("SMTPMessage",returnmsg);
					}
					
					String errormsg = processErrorMsg(returnmsg);
					if(errormsg != null)   {
						System.out.println(errormsg);
						notifyError(errormsg);
						isError = true;
						break;
					}	
				}	
				
				if(!isError)  {
					notifyMessage();
				}	
			}		
		} catch (ConnectionException e)  {
			System.out.println(e.toString());
		}		
	}	
	
	/**
	* According to the response from mail server, return false if the error message
	* can not be managed. Return true if the response from server is success or 
	* the error can be fully managed.<BR>
	* Here's standard return codes from email server<P>
	*  500 Syntax error, command unrecognized <BR>
  *  501 Syntax error in parameters or arguments<BR> 
  *  502 Command not implemented <BR>
  *  503 Bad sequence of commands <BR>
  *  504 Command parameter not implemented <BR>
  *  211 System status, or system help reply <BR>
  *  214 Help message <BR>
  *  220 <domain> Service ready <BR>
  *  221 <domain> Service closing transmission channel <BR>
  *  421 <domain> Service not available, <BR>
  *  250 Requested mail action okay, completed <BR>
  *  251 User not local; will forward to <forward-path> <BR>
  *  450 Requested mail action not taken: mailbox unavailable [E.g., mailbox busy] <BR>
  *  550 Requested action not taken: mailbox unavailable [E.g., mailbox not found, no access] <BR>
  *  451 Requested action aborted: error in processing <BR>
  *  551 User not local; please try <forward-path> <BR>
  *  452 Requested action not taken: insufficient system storage <BR>
  *  552 Requested mail action aborted: exceeded storage allocation <BR>
  *  553 Requested action not taken: mailbox name not allowed [E.g., mailbox syntax incorrect] <BR>
  *  354 Start mail input; end with <CRLF>.<CRLF> <BR>
  *  554 Transaction failed <P>	
	* @param msg Returned message from email server
	*/
	protected String processErrorMsg(String msg)   {
		String retstr = null;
		String retcode = msg.substring(0,3);
		int code = (Integer.valueOf(retcode)).intValue();
		switch(code)   {
			case 500: retstr = "Syntax error, command unrecognized"; break;
			case 501: retstr = "Syntax error in parameters or arguments"; break;
			case 502: retstr = "Command not implemented"; break;
			case 503: retstr = "Bad sequence of commands"; break;
			case 504: retstr = "Command parameter not implemented"; break;
			case 221: retstr = "Service closing transmission channel"; break;
			case 421: retstr = "Service not available"; break;
			case 251: retstr = "User not local; will forward to <forward-path>"; break;
			case 450: retstr = "Requested mail action not taken: mailbox unavailable [E.g., mailbox busy]"; break;
			case 550: retstr = "Requested action not taken: mailbox unavailable [E.g., mailbox not found, no access]"; break;
			case 451: retstr = "Requested action aborted: error in processing"; break;
			case 551: retstr = "User not local; please try <forward-path>"; break;
		case 452: retstr = "Requested action not taken: insufficient system storage"; break;
			case 552: retstr = "Requested mail action aborted: exceeded storage allocation"; break;
			case 553: retstr = "Requested action not taken: mailbox name not allowed [E.g., mailbox syntax incorrect]"; break;
			case 554: retstr = "Transaction failed"; break;
		}
		
		return retstr;	
	}
	
	/**
	  * Notify error if _ipAgentAction(class IPRouterClientAction) is not null
	  * @param msg Message to be notified to _ipAgentAction
	  */
	protected void notifyError(String msg)  {
		if(_ipAgentAction != null)   {
			if(_smtpCon != null)  {
				_smtpCon.setErrorMessage(msg);
			}	
			_ipAgentAction.processMessage("SMTP",_smtpCon);
		}
	} 
	
	/**
	  * Notify message(FTPCon itself) if _ipAgentAction(class IPRouterClientAction) is not null
	  * _ipAgentAction will invoke processMessage(String,Object) where String is "SMTP" and
	  * Object is _smtpCon.
	  */
	protected void notifyMessage()  {
		if(_ipAgentAction != null)   {
			_ipAgentAction.processMessage("SMTP",_smtpCon);
		}
	}
	
}