/**
 * @(#)JavaButton.java	2.2 11/25/96 Rodney Ryan (rer@interport.net)
 *
 * Copyright (c) 1996 Rodney Ryan, All Rights Reserved.
 *
 */

package RouterLayer.AgentClient.Example.RCApplet.extawt;

import java.awt.*;

/**
 * A button class.
 *
 * @author 	Rodney Ryan (rer@interport.net)
 * @version 	2.2 11/25/96
 *
 */
public class JavaButton extends Canvas
{
    /**
    * Event id for no event.
    */
    public static final int NO_EVENT = 0;

    /**
    * The event id variable
    */
    public    int        eventID;

    /**
    * The ButtonFace variable.
    */
    protected ButtonFace face;

    /**
    * Indicates if the button is pressed
    */
    protected boolean    pressed = false;

    /**
    * Indicates if the button has the focus
    */
    protected boolean    focused = false;

    /**
    * Indicates if the button is disabled
    */
    protected boolean    disabled = false;

    /**
    * Constructs a new instance.
    * @param face the ButtonFace
    * @param eventID the event ID
    */
    public JavaButton(ButtonFace face,
                      int        eventID)
    {
        super();
        this.face = face;
        this.eventID = eventID;
    }

    /**
    * Constructs an instance that uses the NullButtonFace
    * and the given event ID.
    * @param eventID the event ID
    */
    public JavaButton(int eventID)
    {
        this(NullButtonFace.instance(), eventID);
    }

    /**
    * Constructs an instance that uses the NullButtonFace and
    * the NO_EVENT event id.
    */
    public JavaButton()
    {
        this(NullButtonFace.instance(), NO_EVENT);
    }

    /**
    * Returns true if the button is disabled.
    */
    public boolean disabled()
    {
        return disabled;
    }

    /**
    * Disables the button.
    */
    public void disable()
    {
        disabled = true;
        repaint();
    }

    /**
    * Enables the button.
    */
    public void enable()
    {
        disabled = false;
        repaint();
    }

    /**
    * Handles the mouseDown event.
    * @param e the mouseDown event
    * @param x the x coordinate of the event
    * @param y the y coordinate of the event
    */
    public boolean mouseDown(Event e,
                             int   x,
                             int   y)
    {
        if (!disabled) {
            pressed = true;
            focused = true;
            repaint();
            return true;
        } else return super.mouseDown(e, x, y);
    }

    /**
    * Handles the mouseUp event.
    * @param e the mouseUp event
    * @param x the x coordinate of the event
    * @param y the y coordinate of the event
    */
    public boolean mouseUp(Event e,
                           int   x,
                           int   y)
    {
        if (!disabled) {
            if ((pressed) && (eventID != NO_EVENT)) notifyParent();
            pressed = false;
            focused = false;
            repaint();
            return true;
        } else return super.mouseUp(e, x, y);
    }

    /**
    * Handles the mouseEnter event.
    * @param e the mouseEnter event
    * @param x the x coordinate of the event
    * @param y the y coordinate of the event
    */
    public boolean mouseEnter(Event e,
                              int   x,
                              int   y)
    {
        if ((focused) && (!disabled)) {
            pressed = true;
            repaint();
            return true;
        } else return super.mouseEnter(e, x, y);
    }

    /**
    * Handles the mouseExit event.
    * @param e the mouseExit event
    * @param x the x coordinate of the event
    * @param y the y coordinate of the event
    */
    public boolean mouseExit(Event e,
                             int   x,
                             int   y)
    {
        if ((focused) && (!disabled)) {
            pressed = false;
            repaint();
            return true;
        } else return super.mouseExit(e, x, y);
    }

    /**
    * Notifies the parent by posting an event with an id of the
    * eventID.
    */
    public void notifyParent()
    {
        postEvent(new Event(getParent(), eventID, this));
    }

    /**
    * Paints the JavaButton.
    * @param g the Graphics
    */
    public void paint(Graphics g)
    {
        g.setColor(getBackground());
        g.fillRect(0, 0, bounds().width, bounds().height);
        if (!pressed) drawOut(g);
            else drawIn(g);
        g.clipRect(2, 2, bounds().width - 4, bounds().height - 4);
        face.paint(g, disabled, pressed);
    }

    /**
    * Draws the JavaButton pressed in.
    * @param g the Graphics
    */
    public void drawIn(Graphics g)
    {
        Rectangle r = bounds();
        g.setColor(Color.gray);
        g.drawRect(0, 0, r.width - 1, r.height - 1);
    }

    /**
    * Draws the JavaButton pressed out.
    * @param g the Graphics
    */
    public void drawOut(Graphics g)
    {
        Rectangle r = bounds();
        g.setColor(Color.black);
        g.drawLine(0, r.height - 1, r.width - 1, r.height - 1);
        g.drawLine(r.width - 1, r.height - 1, r.width - 1, 0);
        g.setColor(Color.gray);
        g.drawLine(1, r.height - 2, r.width - 2, r.height - 2);
        g.drawLine(r.width - 2, r.height - 2, r.width - 2, 1);
        g.drawLine(0, 0, 0, r.height - 2);
        g.drawLine(0, 0, r.width - 2, 0);
        g.setColor(Color.white);
        g.drawLine(1, r.height - 3, 1, 2);
        g.drawLine(1, 1, r.width - 3, 1);
    }

    /**
    * Returns the preferredSize (the minimumSize())
    */
    public Dimension preferredSize()
    {
        return minimumSize();
    }

    /**
    * Returns a minimumSize of the height and width
    * of the button face + a border of 2 pixels.
    */
    public Dimension minimumSize()
    {
        Dimension d = new Dimension(face.size());
        d.height = d.height + 4;
        d.width = d.width + 4;
        return d;
    }
}