/**
 * @(#)TabCanvas.java	2.2 11/25/96 Rodney Ryan (rer@interport.net)
 *
 * Copyright (c) 1996 Rodney Ryan, All Rights Reserved.
 *
 */

package RouterLayer.AgentClient.Example.RCApplet.extawt;

import java.awt.*;
import java.util.Vector;

/**
 * A Canvas to draw the tabs on.
 *
 * @author 	Rodney Ryan (rer@interport.net)
 * @version 	2.2 11/25/96
 *
 */
public class TabCanvas extends Panel
{
    /**
    * Event id for new tab selection
    */
    public static final int TAB_SELECTED        = 6000;

    /**
    * Event id for changing the position of the tabs.
    */
    public static final int POSITIONING_CHANGED = 6001;

    /**
    * Used for double buffering.
    */
    protected Dimension oldSize = new Dimension(0, 0);

    /**
    * Used for double buffering.
    */
    protected Image     masterImage;

    /**
    * Used for double buffering.
    */
    protected Graphics  masterGraphics;

    /**
    * Used for double buffering.
    */
    protected boolean   forcedPaint = false;

    /**
    * Indicates whether the first tab is visible or not.
    */
    protected boolean     firstTabVisible;

    /**
    * Indicates whether the last tab is visible or not.
    */
    protected boolean     lastTabVisible;

    /**
    * Used to calculate the width and height of each tab.
    */
    protected FontMetrics fm;

    /**
    * Holds all the tabs.
    */
    protected Vector      tabs;

    /**
    * References the present selected tab.
    */
    protected Tab         selectedTab;

    /**
    * Keeps track of the index of the first visible tab.
    */
    protected int         tabIndex;

    /**
    * Constructs a new TabCanvas instance.
    * @param firstTabName the name of the first tab
    */
    public TabCanvas(String firstTabName)
    {
        super();
        tabs = new Vector(5, 5);
        tabs.addElement(new Tab(firstTabName));
        selectedTab = (Tab)tabs.elementAt(0);
        selectedTab.select();
        tabIndex = 0;
    }

    /**
    * Calculates the height of the tabs and sets the
    * tab colors.
    */
    public void addNotify()
    {
        super.addNotify();
        Graphics g = getGraphics();
        fm = g.getFontMetrics(Tab.selectFont);
        Tab.setTabHeight(fm.getHeight() + (2 * fm.getDescent()));
        Tab.setForeground(getForeground());
        Tab.setBackground(getBackground());
    }

    /**
    * Adds a tab.
    * @param tabName the name of the new tab
    */
    protected void addTab(String tabName)
    {
        tabs.addElement(new Tab(tabName));
    }

    /**
    * Returns true if the first tab visible is the first
    * tab, i.e. we are scrolled all the way backwards.
    */
    protected boolean atFirstTab()
    {
        return (tabIndex == 0);
    }

    /**
    * Returns true if the first tab visible is the last
    * tab, i.e. we are scrolled all the way forwards.
    */
    protected boolean atLastTab()
    {
        return (tabIndex == tabs.size() - 1);
    }

    /**
    * If we are not at the last tab, scrolls the tabs
    * one forward.
    */
    protected void moveForward()
    {
        if (tabIndex != (tabs.size() - 1)) {
            tabIndex++;
            forcePaint();
        }
    }

    /**
    * If we are not at the first tab, scrolls the tabs
    * one backward.
    */
    protected void moveBack()
    {
        if (tabIndex != 0) {
            tabIndex--;
            forcePaint();
        }
    }

    /**
    * Returns the value of the boolean firstTabVisible.
    */
    protected boolean firstTabVisible()
    {
        return firstTabVisible;
    }

    /**
    * Returns the value of the boolean lastTabVisible.
    */
    protected boolean lastTabVisible()
    {
        return lastTabVisible;
    }

    /**
    * Selects the given tab.
    * @param newSelectedTab the newly selected tab
    */
    protected void selectTab(Tab newSelectedTab)
    {
        selectedTab.deselect();
        selectedTab = newSelectedTab;
        selectedTab.select();
        postEvent(new Event(getParent(), TAB_SELECTED, selectedTab.getName()));
    }

    /**
    * Handles a mouseDown event.
    * @param e the mouseDown event
    */
    public boolean mouseDown(Event e,
                             int   x,
                             int   y)
    {
        Tab tab;

        for (int i = 0; i < tabs.size(); i++) {
            tab = (Tab)tabs.elementAt(i);
            if ((tab.inside(x, y)) && (!tab.selected()))
                selectTab(tab);
        }
        return true;
    }

    /**
    * Immediately re-draws everything.
    */
    public void immediatePaint()
    {
        forcedPaint = true;
        paint(getGraphics());
    }

    /**
    * Calls repaint to re-draw everything.
    */
    public void forcePaint()
    {
        forcedPaint = true;
        repaint();
    }

    /**
    * Calls paint(g) - never called directly.
    */
    public void update(Graphics g)
    {
        paint(g);
    }

    /**
    * Ensures that when we change size the double-buffered
    * image gets updated.
    */
    public synchronized void reshape(int x,
                                     int y,
                                     int width,
                                     int height)
    {
        super.reshape(x, y, width, height);
        repaint();
    }

    /**
    * Paints the TabCanvas.
    * @param g the Graphics
    */
    public void paint(Graphics g)
    {
        boolean sizeNotZero = ((size().width > 0) && (size().height > 0));
        if (sizeNotZero) {
            boolean resized = (oldSize.width != size().width) || (oldSize.height != size().height);
            if ((resized) || (masterImage == null) || (forcedPaint)) {
                if ((resized) || (masterImage == null)) {
                    masterImage = createImage(size().width, size().height);
                    masterGraphics = masterImage.getGraphics();
                }
                oldSize = size();
                forcedPaint = false;
                masterGraphics.setColor(getBackground());
                masterGraphics.fillRect(0, 0, bounds().width, bounds().height);
                updateMasterImage();
                postEvent(new Event(getParent(), POSITIONING_CHANGED, this));
            }
            g.drawImage(masterImage, 0, 0, null);
        }
    }

    /**
    * Updates the double-buffer image.
    */
    protected void updateMasterImage()
    {
        Tab tab;

        firstTabVisible = (tabIndex == 0);
        for (int i = 0; i < tabIndex; i++)
            ((Tab)tabs.elementAt(i)).disable();
        int x = 2;
        int y = bounds().height - Tab.getTabHeight();
        for (int i = tabIndex; i < tabs.size(); i++) {
            tab = ((Tab)tabs.elementAt(i));
            tab.enable();
            tab.resize(fm.stringWidth(tab.getName()) + 10, Tab.getTabHeight());
            tab.move(x, y);
            tab.paint(masterGraphics);
            x += tab.width + 2;
        }
        masterGraphics.setColor(Color.white);
        masterGraphics.drawLine(0, bounds().height - 1, bounds().width, bounds().height - 1);
        paintSelectedTab(masterGraphics);
        if (x > bounds().width) chopLastVisibleTab(masterGraphics);
        lastTabVisible = ((tabIndex == tabs.size() - 1) || (x < bounds().width));
    }

    /**
    * Paints the selected tab.
    * @param g the Graphics
    */
    protected void paintSelectedTab(Graphics g)
    {
        if (selectedTab.enabled()) {
            selectedTab.paint(g);
            g.setColor(getBackground());
            g.drawLine(selectedTab.x, bounds().height - 1, selectedTab.x + selectedTab.width, bounds().height - 1);
        }
    }

    /**
    * Called when the width of the last visible tab exceeds
    * the width of the TabCanvas. Draws a choppy line on the
    * end of the tab.
    * @param g the Graphics
    */
    protected void chopLastVisibleTab(Graphics g)
    {
        int x = bounds().width - 10;
        int y = bounds().height - Tab.getTabHeight();
        if (selectedTab.inside(x, y)) y -= 2;
        g.setColor(getBackground());
        g.fillRect(x, 0, bounds().width, bounds().height - 1);
        boolean clipFirstShadowLine = true;
        while (y < bounds().height - 1) {
            g.setColor(Color.gray);
            g.drawLine(x, y, x, y + 4);
            g.setColor(Color.black);
            if (clipFirstShadowLine) {
                g.drawLine(x + 1, y + 2, x + 1, y + 4);
                clipFirstShadowLine = false;
            } else g.drawLine(x + 1, y, x + 1, y + 4);
            y += 4;
            g.setColor(Color.gray);
            g.drawLine(x + 1, y, x + 1, y + 4);
            g.setColor(Color.black);
            g.drawLine(x + 2, y, x + 2, y + 4);
            y += 4;
        }
        g.setColor(Color.white);
        x++;
        y = bounds().height - 1;
        if (!selectedTab.inside(x, y)) g.drawLine(x - 1, y, x + 1, y);
            else g.drawLine(x, y, bounds().width, y);
    }

    /**
    * Returns the preferredSize, which is a width of 0 and a height at least
    * that of the Tab class plus a 2 pixel top border.
    */
    public Dimension preferredSize()
    {
        return new Dimension(0, Tab.getTabHeight() + 2);
    }

    /**
    * Returns the minimumSize (the preferredSize()).
    */
    public Dimension minimumSize()
    {
        return preferredSize();
    }
}