package RouterLayer.Router;

import Abstract.*;
import RouterLayer.Router.*;
import RouterLayer.AgentClient.*;
import KQMLLayer.*;
import BaseLayer.*;

import java.lang.*;
import java.io.*;
import java.util.*;

/**
* Secure the connection. Check whether the register protocol is satisfied. 
* Since there's no register protocol upon which all the agent communities agree, 
* we are just suggesting the protocol,which is neither sufficient nor theoretically 
* correct. We hope this problem to be resolved in the near future. This is our current 
* register protocol messages<BR>
* At first, upon request, Router will accept() and request identification to the
* connected agent using 'identify-self' performative. Then, the client will send its
* identification using 'whoiam'. If the 'whoiam' information is acceptable, the Router
* will send the 'accepted' message to the client. If not, the Router will disconnect the
* connection. For example:<BR>
* <UL>
* <LI>Agent sends message to the Router<BR>
* (register-agent :sender agent :receiver RouterRegistrar :password xxxxx)<BR>
* <LI>Server send message to the agent<BR>
* (identify-self :sender Router :receiver agent)<BR>
* <LI>Client will response<BR>
* (whoiam :sender client :receiver RouterRegistrar :content (contact-information :host aaa.ggg.com :port 4321 :email aaa@bbb.com) :message-method MessageRouter 
* :description (advertize DesignAgent) :KQML-extensions http://cdr.stanford.edu/newprotocol)<BR>
* <LI>Server will send message<BR>
* <LI>(register-accepted :sender server :receiver client)<BR>
* </UL>
* @see RouterLayer.Router.RegistrarRecvThread
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class RegistrarSecurity extends Security  {

  static final int CLIENTOK = 0;
  static final int CLIENTNAMEMISSING = 1;
  static final int CLIENTSAMENAME = 2;
  static final int PASSWORDMISSING = 3;
  static final int PASSWORDMISMATCH = 4;
  static final int PASSWORDFILECORRUPTED = 5;	
  static final int NOCLIENTLIST = 6;
  static final int NOTVALIDMESSAGEMETHOD = 7;
  static final int REGISTEROK = 8;
  static final int CLIENTALREADYEXIST = 9;
  static final int INTERNALERROR = 100;
  
  /**
   * RouterSecurity to update the user list
   */	
  RouterSecurity _security;
  
  /**
   * Default constructor
   */
  public RegistrarSecurity()  {
    super();
  }
  
  /**
   * Constructor
   * The RouterSecurity is passed as an argument to update the user list after
   * registration is successful.
   * @param security Main Router security
   */
  public RegistrarSecurity(RouterSecurity security)  {
    super(security.getAddressTable(),security.getPasswordFilePath());
    // to enforce security to update the user list
    _security = security;
  }
  
  /**
   * Security check process from the RegistrarServerThread. The connection is accepted and
   * the security check is performed using this method. For error condition, the
   * connection may be denied after sending error message.<P>
   * For both registration and connection, the lowest level connection protocol is:<BR>
   * Router : 201 AMR "Router name"<BR>
   * or
   * Router : 500 Internal Error<BR>
   * for any error conditions.<P>
   * Then, the Agent will request register by sending "register-agent" protocol.<BR>
   * For example,<BR>
   * Agent : (register-agent :sender agent :receiver RouterRegistrar :password xxxxxx)<BR>
   * Router : (identify-self :sender RouterRegistrar :receiver agent)<BR>
   * Agent : (whoiam :sender agent :receiver RouterRegistrar :message-method AMR :KQML-extensions 
   * http://java.stanford.edu/myprotocol :description (advertize DesignAgent) 
   * :content (contact-information :host www.sun.com :port 1234 :email aaa@sun.com))<BR>
   * Router : (register-accepted :sender RouterRegistrar :receiver agent)>BR>
   * The router will disconnect the connection just after the registration process is finished.
   * The agent use "reconnect-agent" protocol to connect the agent, afterwards.
   * For more information for the protocol, refer to <A HREF="http://cdr.stanford.edu/ProcessLink/KQML-proposed.html>AMR protocols</A>
   * @param con RouterRecvThread, will be started
   */
  public void processServerLogin(ReceiverThread con)   {
    
    // cross check will be performed. By sending host address and port
    // the client side will check. If correct the client will send 
    // register message to server
    Address address = con.getAddress();
    
    try  {
      String myid = address.getID();
      
      String loginString = "201 AMR ";
      loginString = loginString + myid;
      con.writeMsg(loginString);
      
      String response = con.getOneMessage();	
      KQMLmessage kqml = new KQMLmessage(response);
      String perf = kqml.getValue("performative");
      perf = perf.toLowerCase();
      
	    if(perf.equals("register-agent"))  {
				String agentname = kqml.getValue("sender");	
				if(agentname == null)  {
	  			// sender not specified. error condition
	  			con.writeMsg("Sender not specified");
	  			return;
				}	
	
				// password is optional
				String password = kqml.getValue("password");
	
				// identify-self protocol
				loginString = "(identify-self :sender ";
				loginString = loginString + address.getID() + " :receiver " + agentname + ")";
				con.writeMsg(loginString);
	
				response = con.getOneMessage();	
				KQMLmessage wkqml = new KQMLmessage(response);
				perf = wkqml.getValue("performative");
	
				if(perf == null)  {
	  			denyConnection(con,myid);
				}
	
				perf = perf.toLowerCase();
	
				if(perf.equals("whoiam"))  {
	  			String agentid = wkqml.getValue("sender");
	  			String content = wkqml.getValue("content");
	  
	  			// currently not used.
	  			String kqmlextensions = wkqml.getValue("KQML-extensions");
	  			String type = wkqml.getValue("message-method");
	  			String description = wkqml.getValue("description");
	  
	  			// these will be in the content field.
	  			String host = null;
	  			String port = null;
	  			String email = null;
	  			if(content != null)  {
	    			try  {
	      			// the description field will include password
	      			KQMLmessage infokqml = new KQMLmessage(content);
	      
	      			// can be 'null' if the agent is not a stand-alone agent
	      			host = infokqml.getValue("host");
	      
	      			// can be -1 if the agent is not a stand-alone agent
	      			port = infokqml.getValue("port");
							// optional
	      			email = infokqml.getValue("email");
	    			} catch (ParseException p)  {
	      			con.writeMsg("Contact-information parse error");
	      			return;
	    			}
	  			}		
	  			else  {
	    			con.writeMsg("Contact-information is missing");
	    			return;
	  			}		
	  
	  			// make new internal description field
				  String newdescription = "(agent-info";
	  			if(kqmlextensions != null)  {
	    			newdescription = newdescription + " :KQML-extensions " + kqmlextensions;
	  			}
	  			if(description != null)  {
	    			newdescription = newdescription + " :description " + description;
	  			}
	  			if(email != null)  {
	    			newdescription = newdescription + " :email " + email;
	  			}
	  			if(password != null)  {
	    			newdescription = newdescription + " :password " + password;		
	  			}	
	  			newdescription = newdescription + ")";
	  
	  			int checkresult = secureAgent(myid,agentid);
	 			 if(checkresult == CLIENTOK)  {
	    
	 			   // check the agent name is unique. If password check is enforced, password should
	 			   // not be null			 
	 			   int errorcode = registerAgent(agentid,host,port,type,newdescription,password);
	    
	 			   if(errorcode != REGISTEROK)  {
	 			     String error = "(error :sender " + myid + " :receiver " + agentid + " :content (";
	 			     switch(errorcode)  {
	 			     case NOTVALIDMESSAGEMETHOD: error = error + "invalid-message-method))";
							break;
				     case PASSWORDMISSING: error = error + "password-missing))";
							break;
				     case INTERNALERROR: error = error + "internal-error))";
							break;
	      		 case PASSWORDMISMATCH: error = error + "password-mismatch))";    
	      	 }
	      	 con.writeMsg(error);
	      	 denyConnection(con,myid);	
	    	}			
	    	else  {		
	      	startConnection(con,agentid,myid);
	    	}		
	  }	
	  else  {
	    processErrorLogin(checkresult,con,myid);		    		
	  }
	} 	
	else  {
	  denyConnection(con,myid);
	}	
      }
      else  {
				denyConnection(con,myid);
      }	
    }  catch (Exception e)  {
      System.out.println(e.toString());
    }
  }					
  
  /**
   * According to the error condition, the processErrorLogin will be performed.
   * If the agent name is missing or if the agent name is the same as the Router
   * name or if the password is missing(when passwordCheckBit is true) or 
   * if the password is mismatch(when passwordCheckBit is true), the Registrar will
   * send 'error' message to the agent and will close the connection
   * @param errorCode Error code which happens during connection process
   * @param con ReceiverRecvThread, which wants to be connected
   * @param myid The Router name
   */
  protected boolean processErrorLogin(int errorCode,ReceiverThread con,String myid)   {		  
    try  {
      String error = "(error :sender ";
      error = error + myid + " :content (";
      
      switch(errorCode)   {
      case CLIENTNAMEMISSING:
	error = error + " name-wrong)";
	break;
      case CLIENTSAMENAME:
	error = error + " name-duplicated)";
	break;
      case PASSWORDMISSING:
	error = error + " password-missing)";
	break;
      case CLIENTALREADYEXIST:
	error = error + " already-registered)";	
	break;
      default:
	error = error + ")";	
      }		
      error = error + ")";
      // send error
      con.writeMsg(error);
    }		catch (Exception e)  {
      System.out.println(e.toString());
    }
    return false;
  }				
  
  /**
   * Start conneciton. This does not mean that the connection will be established.
   * After register, the agent should reconnect to the Router. This method will
   * send 'register-accepted' message to the client and will be kill the RegistrarRecvThread
   * @param con ReceiverThread to be started 
   * @param agentid The agent name
   * @param myid The Router registrar name
   */
  protected void startConnection(ReceiverThread con,String agentid,String myid)  {
    System.out.println("Register Request from " + agentid + " accepted");				
    String access = "(register-accepted :sender ";
    access = access + myid + " :receiver " + agentid + ")";
    try {
      con.writeMsg(access);
    }  catch (ConnectionException e) {}
  } 	
  
  /**
   * Error occurs from the connection process. Send 'error' message to the agent
   * and disconnect the conneciton
   * @param con ReceiverThread to be disconnected
   * @param id The RouterRegistrar name
   */
  protected void denyConnection(ReceiverThread con,String id)  {
    String errormsg = "(error :sender ";
    errormsg = errormsg + id + " :content register-expected)";
    System.out.println("Register Request from " + id + " failed:Not compatible protocol");
    try {
      con.writeMsg(errormsg);
    }  catch (ConnectionException e) {}
  }
  
  /**
   * Check the client send its name and the name is different from mine
   * @param me My name, sent back from the client
   * @param client The agent name
   */
  protected int secureAgent(String me,String client)  {
    if(client == null) return CLIENTNAMEMISSING;
    // client can not have the same name with me
    if(client.equals(me)) return CLIENTSAMENAME;
    
    return CLIENTOK;
  }	
  
  /**
   * Opposite to the processServerLogin. The agent(me) wants to connect to the
   * Router and it will send 'whoiam' message to the Router
   * @param c ReceiverThread, which will connect to the Router
   * @param serveraddress The Rotuer server address
   */
  public void processClientLogin(ReceiverThread c,Address serveraddress)  
  	throws ConnectionException   {
    if(_addresses == null)  {
      SystemOut("No AddressTable is defined");
      throw new ConnectionException("No AddressTable is defined");
    }
    
    // My address should be specified
    Address myaddress = c.getAddress();
    if(myaddress == null)   {
      SystemOut("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
      throw new ConnectionException("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
    }
    
    if(serveraddress == null)   {
      SystemOut("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
      throw new ConnectionException("Address for " + c.getAddress().getID() + " is not specified. Fail to client connect");
    }  		
    
    String initial = c.getOneMessage();
    try  {
      if(!initial.startsWith("201 AMR"))  {
      	throw new ConnectionException("Port is not for AMR Registrar");
      }
      
      System.out.println("Start to register");
  		
      // description field will look like
      // (agent-info :KQML-extensions URL :description (advertize designAgent) :email a@bbb.com
      // :password xxxxxxx)	
      String description = myaddress.getDescription();
      if(description != null)  {
	KQMLmessage dkqml = new KQMLmessage(description);
	String kqmlextensions = dkqml.getValue("KQML-extensions");
	String describe = dkqml.getValue("description");
	String email = dkqml.getValue("email");
	String password = dkqml.getValue("password");
	
	String register = "(register-agent :sender ";
	register = register + myaddress.getID();
	if(password != null)  {
	  register = register + " :password " + password;
	}
	register = register + ")";
	
	c.writeMsg(register);	
	
	initial = c.getOneMessage();
	
	// register protocol 'identify'
	KQMLmessage kqml = new KQMLmessage(initial);
  	
	String perf = kqml.getValue("performative");
	if(perf.equals("identify-self"))  {
	  String content = "(contact-information :host " + myaddress.getHost() + 
	    " :port " + Integer.toString(myaddress.getPort());
	  if(email != null)  {
	    content = content + " :email " + email;
	  }
	  content = content + ")";	
	  
	  
	  String loginString = "(whoiam :sender ";
	  loginString = loginString + myaddress.getID() + " :message-method " + myaddress.getType();
	  loginString = loginString + " :receiver " + serveraddress.getID();
	  loginString = loginString + " :content " + content;
	  
	  if(kqmlextensions != null)  {
	    loginString = loginString + " :KQML-extensions " + kqmlextensions;
	  }
	  loginString = loginString + ")";
	  
	  c.writeMsg(loginString);  		
	  String response = c.getOneMessage();
	  
	  
	  KQMLmessage respkqml = new KQMLmessage(response);
	  String respperf = respkqml.getValue("performative");
	  if(respperf.equals("register-accepted"))  {		
	    SystemOut("Register accepted");
	    return;
	  }
	  else  {
	    SystemOut("Connection need more information::" + response);
	    throw new ConnectionException("Registrer failed " + response);
	  }		
	}		
	else
	  if(perf.equals("error"))  {
	    SystemOut("Connection failed for " + c.getAddress().getID());
	    throw new ConnectionException("Registrar throw error " + kqml.getReadString());
	  }
	  else  {
	    SystemOut("Incompatable protocol");
	    throw new ConnectionException("Not AMR register protocol " + kqml.getReadString());
	  }		
      }
      else  {
	SystemOut("Description field should be specified");
      	throw new ConnectionException("Description Field is not specified. Register failed");
      }	
      
    } catch (Exception e)  {
      SystemOut(e.toString());    
      // IOException or KQML parse exception. Not allowed
      throw new ConnectionException("Registrar throw error " + e.toString());
    }		
  }
  
  /**
   * Registrar does not send out disconnect message
   * @param c RouterRecvThread to be disconnected
   */
  public void processClientLogout(ReceiverThread c)  {
    // do nothing. Automatic disconnect
  }	
  
  /**
   * Register the agent to the agent list. The address table will be updated
   * @param id Agent name
   * @param host Internet host name
   * @param port Port number
   * @param type Message method
   * @param description Agent description
   * @param password Password for the agent to be checked afterwards
   */
  public int registerAgent(String id,String host,String port,String type,String description,String password)  {
    try  {
      if(type == null)  return NOTVALIDMESSAGEMETHOD;
      if(password == null) return PASSWORDMISSING;
      
      String[] sdata = new String[2];
      sdata[0] = id;
      sdata[1] = password;
      if(!isValidAgent(sdata))  {
      	return PASSWORDMISMATCH;
      }
      
      Address addr = new Address(id,host,(Integer.valueOf(port)).intValue(),type,description);
      
      _security.updateUserList(addr,password);
      
//      _addresses.addAddress(addr); 	
      
      return REGISTEROK;
    } catch (Exception e)  {
      System.out.println(e.toString());
    }
    
    return INTERNALERROR;		
  }	
  
  /**
   * Check the user is valid
   * @param o String argument, agent name and password
   * @return True if the agent is valid
   */
  public boolean isValidAgent(Object o)  {
    if(_security != null)  {
      String[] args = (String[])o;
      if(_security.isRegisteredUser(args[0]))  {
	return _security.isValidAgent(o);
      }
    }	
    return true;
    // you may need something here
  }  
  
  /**
   * If you use GUI, override this method
   * @param str String which will be displayed to the GUI or System.out
   */
  public void SystemOut(String str)  {
    System.out.println(str);
  }
  
}		