package RouterLayer.Router;

import Abstract.*;
import BaseLayer.*;

import java.net.*;
import java.lang.*;
import java.io.*;
import java.util.*;
/**
* Router server thread. 
* Two data members, incoming message box directiry and reservation table are added.
* The run() method will accept the connection from the Router clients and create
* RouterRecvThread.
*/

/*
* Copyright (c) 1996,1997,1998 <A HREF="http://cdr.stanford.edu/">Center for Design Research</A>, Stanford University.
* All rights reserved.<p>
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You can obtain the GNU General Public License at
* http://www.gnu.org/copyleft/gpl.html or by writing to the Free
* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
* 
* This software is bound by the terms and conditions listed in the
* attached <a href="../LICENSE">LICENSE file</A>.
*  @author <A HREF="http://cdr.stanford.edu/html/people/jeon-bio.html>Heecheol Jeon</A>
*  @version Java(tm) 1.1
*/ 

public class RouterServerThread extends BServerThread  {
	
	/**
	* Incoming message box directory
	*/
	protected String _messageBoxPath;
	
	/**
	* Message reservation table. The key field is time stamp and field value is 
	* Reservation object
	* @see RouterLayer.Router.Reservation
	*/
	protected Hashtable _reservedBook;
	
	/**
	* Agent name array which contains the agent names those are off-line intentionally.
	* If a message receiver is in the off-line list, the RouterRecvThread will not send
	* the message to the receiver. The message will be saved to the file. If the receiver
	* is reconnect, the received messages will be sent.
	*/
	protected OffLineAgentTable _offLineList;
	
	/**
	* Vector which contains the disconnected agent name list
	*/
	protected Vector _disconnectTable;
	
	
  /**
  * Set the ReservationTable
  * @param reservedBook ReservationTable from the RouterAction
  */
  public void setReservedBook(Hashtable reservedBook)  {
  	_reservedBook = reservedBook;
  }	
  
  /**
  * Set the off line agent name list. The off-line means that an agent
  * intentionally wants not to receive message before it reconnect to the Router.
  * @param offlinelist Off line agent name list. 
  */
  public void setOffLineList(OffLineAgentTable offlinelist)  {
  	_offLineList = offlinelist;
  }
	
	
	/**
	* Default Constructor
	*/
	public RouterServerThread()  {
	}
	
	/**
	* Constructor
	* @param myaddress Router server address
	* @param connections ConnectionTable, which will be passed to the RouterRecvThread
	* @param security RouterSecurity
	* @param messageboxpath Incoming message box directory
	* @param reservedbook Message reservation table
	* @param offlinelist The off line agent name list.
	* @param disconnectlist The disconnected agents name list.
	* @exception IOException if socket error occurs
	* @exception ConnectionException if connection related error occurs
	* @exception UnknownHostException if host address is not found
	*/
	public RouterServerThread(Address myaddress,
														ConnectionTable connections,
														Security security,
														String messageboxpath,
														Hashtable reservedBook,
														OffLineAgentTable offlinelist,
														Vector disconnectlist) 
	throws IOException,ConnectionException,UnknownHostException {
		super(myaddress,
				  connections,
				  null,
				  security);
		_messageBoxPath = messageboxpath;
		_reservedBook = reservedBook;
		_offLineList = offlinelist;
		_disconnectTable = disconnectlist;
  }
	/**
	* Run method will create RouterRecvThread by accepting connection.
	* endWith character and maximum idle time will be set to the RouterRecvThread
	*/
	public void run()  {
    while (true) {
    	Socket client_socket = null;
	    try {
				client_socket = _listenSocket.accept();
			} catch (IOException e)  {
				// try to recreate server socket if fails	
				try  {
					if(_listenSocket != null)  {
						_listenSocket.close();
					}
		
					_listenSocket = new ServerSocket(_myaddress.getPort());
				} catch (Exception ee)  {
					System.out.println(ee.toString());
				}
			}	
				
			try  {	
				if(client_socket != null)  {
					createReceiverThread(client_socket);	
				}	
    	}  catch (Exception eeee) {
    		// just ignore
    		System.out.println(eeee.toString());
    	}	
    }	
  } 
 
 	/**
 	* Create RouterReceiverThread and invoke start() method
 	* @param socket Socket, created by the ServerSocket
 	*/
  protected synchronized void createReceiverThread(Socket socket)  {
  	try  {
			RouterRecvThread c = new RouterRecvThread(socket,
									Thread.NORM_PRIORITY,_myaddress,_connections,null,_security,_messageBoxPath,_reservedBook,_offLineList,_disconnectTable);
									
			// do not forget to set end with character if it is not '\004'					
			c.setEndWith(_endWith);					
			if(_durationTime != -1)  {
				c.setDurationTime(_durationTime);
			}	
				
			c.start();
		} catch (ConnectionException c)  {
		}	
  } 	
 	

}				  	